import logging
from typing import Optional, List, AsyncGenerator, Tuple, Any, Dict
from royalnet.bard.discord import YtdlDiscord
from royalnet.serf.discord import Playable
import discord


log = logging.getLogger(__name__)


class RoyalQueue(Playable):
    """A queue of :class:`YtdlDiscord` to be played in sequence."""
    def __init__(self, start_with: Optional[List[YtdlDiscord]] = None):
        super().__init__()
        self.contents: List[YtdlDiscord] = []
        self.now_playing: Optional[YtdlDiscord] = None
        if start_with is not None:
            self.contents = [*self.contents, *start_with]
        log.debug(f"Created new {self.__class__.__qualname__} containing: {self.contents}")

    async def _generator(self) \
            -> AsyncGenerator[Optional["discord.AudioSource"], Tuple[Tuple[Any, ...], Dict[str, Any]]]:
        yield
        while True:
            log.debug(f"Dequeuing an item...")
            try:
                # Try to get the first YtdlDiscord of the queue
                self.now_playing: YtdlDiscord = self.contents.pop(0)
            except IndexError:
                # If there isn't anything, yield None
                log.debug(f"Nothing to dequeue, yielding None.")
                yield None
                continue
            log.debug(f"Yielding FileAudioSource from: {self.now_playing}")
            # Create a FileAudioSource from the YtdlDiscord
            # If the file hasn't been fetched / downloaded / converted yet, it will do so before yielding
            async with self.now_playing.spawn_audiosource() as fas:
                # Yield the resulting AudioSource
                yield fas
            # Delete the YtdlDiscord file
            log.debug(f"Deleting: {self.now_playing}")
            await self.now_playing.delete_asap()
            log.debug(f"Deleting: {self.now_playing.ytdl_file}")
            await self.now_playing.ytdl_file.delete_asap()
            log.debug(f"Deleted successfully!")
            self.now_playing = None



from typing import *
from royalnet.constellation.api import *
from royalnet.backpack.tables.users import User
from royalnet.utils import asyncify

async def find_user_api(input: Union[int, str], alchemy, session):
    if isinstance(input, int):
        user_id = input
    elif isinstance(input, str):
        try:
            user_id = int(input)
        except ValueError:
            raise InvalidParameterError(f"Invalid user id passed to {find_user_api.__name__}")
    else:
        raise TypeError(f"Invalid input type passed to {find_user_api.__name__}")
    user: User = await asyncify(session.query(alchemy.get(User)).get, user_id)
    if user is None:
        raise NotFoundError("No such user.")
    return user



from sqlalchemy import *
from sqlalchemy.orm import *
from sqlalchemy.ext.declarative import declared_attr


class FiorygiTransaction:
    __tablename__ = "fiorygitransactions"

    @declared_attr
    def id(self):
        return Column(Integer, primary_key=True)

    @declared_attr
    def change(self):
        return Column(Integer, nullable=False)

    @declared_attr
    def user_id(self):
        return Column(Integer, ForeignKey("fiorygi.user_id"), nullable=False)

    @declared_attr
    def wallet(self):
        return relationship("Fiorygi", backref=backref("transactions"))

    @property
    def user(self):
        return self.wallet.user

    @declared_attr
    def reason(self):
        return Column(String, nullable=False, default="")

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.change:+} to {self.user.username} for {self.reason}>"



import discord
from typing import *
from royalnet.serf.discord.discordserf import DiscordSerf
from royalnet.commands import *


class DiscordCvEvent(Event):
    name = "discord_cv"

    async def run(self, guild_id: Optional[int] = None, **kwargs) -> dict:
        if not self.interface.name == "discord":
            raise UnsupportedError()

        # noinspection PyTypeChecker
        serf: DiscordSerf = self.interface.serf

        client: discord.Client = serf.client

        if guild_id is None:
            guilds: List[discord.Guild] = client.guilds
            if len(guilds) == 0:
                raise ConfigurationError("Il bot non è in nessun Server.")
            elif len(guilds) > 1:
                raise UserError("Non hai specificato di quale Server vuoi vedere le informazioni!")
            else:
                guild = guilds[0]
        else:
            guild: discord.Guild = client.get_guild(guild_id)

        members: List[Union[discord.User, discord.Member]] = guild.members
        channels: List[discord.VoiceChannel] = guild.voice_channels

        results = []

        for member in members:
            data = {
                "id": member.id,
                "name": member.name,
                "discriminator": member.discriminator,
                "nick": member.nick,
                "bot": member.bot,
                "voice": {
                    "channel": {
                        "id": member.voice.channel.id,
                        "name": member.voice.channel.name,
                        "position": member.voice.channel.position,
                        "category": {
                            "id": member.voice.channel.category_id,
                            "name": member.voice.channel.category.name,
                            "position": member.voice.channel.category.position,
                        } if member.voice.channel.category is not None else {
                            "id": None,
                            "name": None,
                            "position": -1,
                        },
                        "bitrate": member.voice.channel.bitrate,
                        "user_limit": member.voice.channel.user_limit,
                    },
                    "server_mute": member.voice.mute,
                    "server_deaf": member.voice.deaf,
                    "self_mute": member.voice.self_mute,
                    "self_deaf": member.voice.self_deaf,
                    "video": member.voice.self_video,
                    "golive": member.voice.self_stream,
                    "afk": member.voice.afk,
                } if member.voice is not None else None,
                "status": {
                    "main": member.status.value,
                    "desktop": member.desktop_status.value,
                    "mobile": member.mobile_status.value,
                    "web": member.web_status.value,
                },
                "activities": [activity.to_dict() for activity in member.activities if activity is not None],
                "roles": [{
                    "id": role.id,
                    "name": role.name,
                    "hoist": role.hoist,
                    "position": role.position,
                    "managed": role.managed,
                    "mentionable": role.mentionable,
                } for role in member.roles]
            }

            results.append(data)

        return {
            "guild": {
                "id": guild.id,
                "name": guild.name,
                "members": results,
            }
        }



from .play import PlayCommand


class SoundcloudCommand(PlayCommand):
    name: str = "soundcloud"

    aliases = ["sc"]

    description: str = "Cerca un video su SoundCloud e lo aggiunge alla coda della chat vocale."

    syntax = "{ricerca}"

    async def get_url(self, args):
        return f"scsearch:{args.joined()}"



import enum


class DotaMedal(enum.Enum):
    HERALD = 1
    GUARDIAN = 2
    CRUSADER = 3
    ARCHON = 4
    LEGEND = 5
    ANCIENT = 6
    DIVINE = 7
    IMMORTAL = 8

    def __str__(self):
        return self.name.capitalize()

    def __repr__(self):
        return f"{self.__class__.__qualname__}.{self.name}"

    def __gt__(self, other):
        return self.value > other.value



import datetime
from typing import *

import discord
import royalnet.commands as rc
import royalnet.serf.discord as rsd
import royalnet.bard.discord as rbd

from ..utils import RoyalQueue


class DiscordPlayEvent(rc.Event):
    name = "discord_play"

    async def run(self,
                  url: str,
                  guild_id: Optional[int] = None,
                  user: Optional[str] = None,
                  force_color: Optional[int] = None,
                  **kwargs) -> dict:
        if not isinstance(self.serf, rsd.DiscordSerf):
            raise rc.UnsupportedError()

        serf: rsd.DiscordSerf = self.serf
        client: discord.Client = self.serf.client

        # TODO: fix this in Royalnet sometime
        candidate_players: List[rsd.VoicePlayer] = []
        for player in serf.voice_players:
            player: rsd.VoicePlayer
            if not player.voice_client.is_connected():
                continue
            if guild_id is not None:
                guild = client.get_guild(guild_id)
                if guild != player.voice_client.guild:
                    continue
            candidate_players.append(player)

        if len(candidate_players) == 0:
            raise rc.UserError("Il bot non è in nessun canale vocale.\n"
                               "Evocalo prima con [c]summon[/c]!")
        elif len(candidate_players) == 1:
            voice_player = candidate_players[0]
        else:
            raise rc.CommandError("Non so in che Server riprodurre questo file...\n"
                                  "Invia il comando su Discord, per favore!")

        ytds = await rbd.YtdlDiscord.from_url(url)
        added: List[rbd.YtdlDiscord] = []
        too_long: List[rbd.YtdlDiscord] = []
        if isinstance(voice_player.playing, RoyalQueue):
            for index, ytd in enumerate(ytds):
                if ytd.info.duration >= datetime.timedelta(seconds=self.config["Play"]["max_song_duration"]):
                    too_long.append(ytd)
                    continue
                await ytd.convert_to_pcm()
                added.append(ytd)
                voice_player.playing.contents.append(ytd)
            if not voice_player.voice_client.is_playing():
                await voice_player.start()
        else:
            raise rc.CommandError(f"Non so come aggiungere musica a [c]{voice_player.playing.__class__.__qualname__}[/c]!")

        main_channel: discord.TextChannel = client.get_channel(self.config["Discord"]["main_channel_id"])

        if len(added) > 0:
            if user:
                await main_channel.send(rsd.escape(f"▶️ {user} ha aggiunto {len(added)} file alla coda:"))
            else:
                await main_channel.send(rsd.escape(f"▶️ Aggiunt{'o' if len(added) == 1 else 'i'} {len(added)} file alla"
                                                   f" coda:"))
        for ytd in added:
            embed: discord.Embed = ytd.embed()
            if force_color:
                embed._colour = discord.Colour(force_color)
            await main_channel.send(embed=embed)

        if len(too_long) > 0:
            if user:
                await main_channel.send(rsd.escape(
                    f"⚠ {len(too_long)} file non {'è' if len(too_long) == 1 else 'sono'}"
                    f" stat{'o' if len(too_long) == 1 else 'i'} scaricat{'o' if len(too_long) == 1 else 'i'}"
                    f" perchè durava{'' if len(too_long) == 1 else 'no'}"
                    f" più di [c]{self.config['Play']['max_song_duration']}[/c] secondi."
                ))

        if len(added) + len(too_long) == 0:
            raise

        return {
            "added": [{
                "title": ytd.info.title,
            } for ytd in added],
            "too_long": [{
                "title": ytd.info.title,
            } for ytd in too_long]
        }



import discord
from typing import *
from royalnet.commands import *
from royalnet.serf.discord import *


class DiscordSkipEvent(Event):
    name = "discord_skip"

    async def run(self,
                  guild_id: Optional[int] = None,
                  **kwargs) -> dict:
        if not isinstance(self.serf, DiscordSerf):
            raise UnsupportedError()
        client: discord.Client = self.serf.client
        if len(self.serf.voice_players) == 1:
            voice_player: VoicePlayer = self.serf.voice_players[0]
        else:
            if guild_id is None:
                # TODO: trovare un modo per riprodurre canzoni su più server da Telegram
                raise InvalidInputError("Non so in che Server riprodurre questo file...\n"
                                        "Invia il comando su Discord, per favore!")
            guild: discord.Guild = client.get_guild(guild_id)
            if guild is None:
                raise InvalidInputError("Impossibile trovare il Server specificato.")
            voice_player: VoicePlayer = self.serf.find_voice_player(guild)
            if voice_player is None:
                raise UserError("Il bot non è in nessun canale vocale.\n"
                                "Evocalo prima con [c]summon[/c]!")
        # Stop the playback of the current song
        voice_player.voice_client.stop()
        # Done!
        return {}



import asyncio
import logging
import sentry_sdk
import aiohttp
from typing import *
from royalnet.commands import *
from royalnet.utils import *
from royalnet.serf.telegram.escape import escape as tg_escape
from ..tables import Steam, Dota
from ..types import DotaRank

log = logging.getLogger(__name__)


class DotaCommand(Command):
    name: str = "dota"

    aliases = ["dota2", "doto", "doto2", "dotka", "dotka2"]

    description: str = "Visualizza le tue statistiche di Dota!"

    syntax: str = ""

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        if self.interface.name == "telegram":
            self.loop.create_task(self._updater(900))

    async def _send(self, message):
        client = self.serf.client
        await self.serf.api_call(client.send_message,
                                 chat_id=self.config["Telegram"]["main_group_id"],
                                 text=tg_escape(message),
                                 parse_mode="HTML",
                                 disable_webpage_preview=True)

    @staticmethod
    def _display(dota: Dota) -> str:
        string = f"ℹ️ [b]{dota.steam.persona_name}[/b]\n"

        if dota.rank:
            string += f"{dota.rank}\n"

        string += f"\n" \
                  f"Wins: [b]{dota.wins}[/b]\n" \
                  f"Losses: [b]{dota.losses}[/b]\n" \
                  f"\n"

        return string

    async def _notify(self,
                      obj: Dota,
                      attribute_name: str,
                      old_value: Any,
                      new_value: Any):
        if attribute_name == "wins":
            if old_value is None:
                message = f"↔️ Account {obj} connesso a {obj.steam.user}!"
                await self._send(message)
        elif attribute_name == "rank":
            old_rank: Optional[DotaRank] = old_value
            new_rank: Optional[DotaRank] = new_value
            if new_rank > old_rank:
                message = f"📈 [b]{obj.steam.user}[/b] è salito a [b]{new_value}[/b] su Dota 2! Congratulazioni!"
            elif new_rank < old_rank:
                message = f"📉 [b]{obj.steam.user}[/b] è sceso a [b]{new_value}[/b] su Dota 2."
            else:
                return
            await self._send(message)

    @staticmethod
    async def _change(obj: Dota,
                      attribute_name: str,
                      new_value: Any,
                      callback: Callable[[Dota, str, Any, Any], Awaitable[None]]):
        old_value = obj.__getattribute__(attribute_name)
        if old_value != new_value:
            await callback(obj, attribute_name, old_value, new_value)
        obj.__setattr__(attribute_name, new_value)

    async def _update(self, steam: Steam, db_session):
        log.info(f"Updating: {steam}")
        log.debug(f"Getting player data from OpenDota...")
        async with aiohttp.ClientSession() as session:
            # Get profile data
            async with session.get(f"https://api.opendota.com/api/players/{steam.steamid.as_32}/") as response:
                if response.status != 200:
                    raise ExternalError(f"OpenDota / returned {response.status}!")
                p = await response.json()
                # No such user
                if "profile" not in p:
                    log.debug(f"Not found: {steam}")
                    return
            # Get win/loss data
            async with session.get(f"https://api.opendota.com/api/players/{steam.steamid.as_32}/wl") as response:
                if response.status != 200:
                    raise ExternalError(f"OpenDota /wl returned {response.status}!")
                wl = await response.json()
                # No such user
                if wl["win"] == 0 and wl["lose"] == 0:
                    log.debug(f"Not found: {steam}")
                    return
        # Find the Dota record, if it exists
        dota: Dota = steam.dota
        if dota is None:
            dota = self.alchemy.get(Dota)(steam=steam)
            db_session.add(dota)
            db_session.flush()
        await self._change(dota, "wins", wl["win"], self._notify)
        await self._change(dota, "losses", wl["lose"], self._notify)
        if p["rank_tier"]:
            await self._change(dota, "rank", DotaRank(rank_tier=p["rank_tier"]), self._notify)
        else:
            await self._change(dota, "rank", None, self._notify)

    async def _updater(self, period: int):
        log.info(f"Started updater with {period}s period")
        while True:
            log.info(f"Updating...")
            session = self.alchemy.Session()
            log.info("")
            steams = session.query(self.alchemy.get(Steam)).all()
            for steam in steams:
                try:
                    await self._update(steam, session)
                except Exception as e:
                    sentry_sdk.capture_exception(e)
                    log.error(f"Error while updating {steam.user.username}: {e}")
                await asyncio.sleep(1)
            await asyncify(session.commit)
            session.close()
            log.info(f"Sleeping for {period}s")
            await asyncio.sleep(period)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        author = await data.get_author(error_if_none=True)

        found_something = False

        message = ""
        for steam in author.steam:
            await self._update(steam, data.session)
            if steam.dota is None:
                continue
            found_something = True
            message += self._display(steam.dota)
            message += "\n"
        if not found_something:
            raise UserError("Nessun account di Dota 2 trovato.")
        await data.reply(message)



import pickle
import base64
import discord
from typing import *
from royalnet.commands import *
from royalnet.utils import *


class QueueCommand(Command):
    name: str = "queue"

    aliases = ["q"]

    description: str = "Visualizza la coda di riproduzione attuale.."

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        if self.interface.name == "discord":
            message: discord.Message = data.message
            guild: discord.Guild = message.guild
            guild_id: Optional[int] = guild.id
        else:
            guild_id = None
        response: Dict[str, Any] = await self.interface.call_herald_event("discord", "discord_queue",
                                                                          guild_id=guild_id)

        queue_type = response["type"]
        if queue_type == "RoyalQueue":
            next_up = response["next_up"]
            now_playing = response["now_playing"]
            await data.reply(f"ℹ️ La coda contiene {len(next_up)} file.\n\n")

            if now_playing is not None:
                reply = f"Attualmente, sta venendo riprodotto:\n"
                if self.interface.name == "discord":
                    await data.reply(reply)
                    embed = pickle.loads(base64.b64decode(bytes(now_playing["stringified_base64_pickled_discord_embed"],
                                                                encoding="ascii")))
                    # noinspection PyUnboundLocalVariable
                    await message.channel.send(embed=embed)
                else:
                    reply += f"▶️ {now_playing['title']}\n\n"
                    await data.reply(reply)
            else:
                await data.reply("⏹ Attualmente, non sta venendo riprodotto nulla.")

            reply = ""
            if len(next_up) >= 1:
                reply += "I prossimi file in coda sono:\n"
                if self.interface.name == "discord":
                    await data.reply(reply)
                    for item in next_up[:5]:
                        embed = pickle.loads(base64.b64decode(bytes(item["stringified_base64_pickled_discord_embed"],
                                                                    encoding="ascii")))
                        # noinspection PyUnboundLocalVariable
                        await message.channel.send(embed=embed)
                else:
                    reply += numberemojiformat([a["title"] for a in next_up[:5]])
                    await data.reply(reply)
            else:
                await data.reply("ℹ️ Non ci sono altri file in coda.")
        else:
            raise CommandError(f"Non so come visualizzare il contenuto di un [c]{queue_type}[/c].")



from .play import PlayCommand


class GooglevideoCommand(PlayCommand):
    name: str = "googlevideo"

    aliases = ["gv"]

    description: str = "Cerca un video su Google Video e lo aggiunge alla coda della chat vocale."

    syntax = "{ricerca}"

    async def get_url(self, args):
        return f"gvsearch:{args.joined()}"

    # Too bad gvsearch: always finds nothing.



import discord
import pickle
import base64
from typing import *
from royalnet.commands import *
from royalnet.serf.discord import *
from ..utils import RoyalQueue


class DiscordQueueEvent(Event):
    name = "discord_queue"

    async def run(self,
                  guild_id: Optional[int] = None,
                  **kwargs) -> dict:
        if not isinstance(self.serf, DiscordSerf):
            raise UnsupportedError()
        client: discord.Client = self.serf.client
        if len(self.serf.voice_players) == 1:
            voice_player: VoicePlayer = self.serf.voice_players[0]
        else:
            if guild_id is None:
                # TODO: trovare un modo per riprodurre canzoni su più server da Telegram
                raise InvalidInputError("Non so in che Server riprodurre questo file...\n"
                                        "Invia il comando su Discord, per favore!")
            guild: discord.Guild = client.get_guild(guild_id)
            if guild is None:
                raise InvalidInputError("Impossibile trovare il Server specificato.")
            voice_player: VoicePlayer = self.serf.find_voice_player(guild)
            if voice_player is None:
                raise UserError("Il bot non è in nessun canale vocale.\n"
                                "Evocalo prima con [c]summon[/c]!")
        if isinstance(voice_player.playing, RoyalQueue):
            now_playing = voice_player.playing.now_playing
            return {
                "type": f"{voice_player.playing.__class__.__qualname__}",
                "now_playing": {
                    "title": now_playing.info.title,
                    "stringified_base64_pickled_discord_embed": str(base64.b64encode(pickle.dumps(now_playing.embed())),
                                                                    encoding="ascii")
                } if now_playing is not None else None,
                "next_up": [{
                    "title": ytd.info.title,
                    "stringified_base64_pickled_discord_embed": str(base64.b64encode(pickle.dumps(ytd.embed())),
                                                                    encoding="ascii")
                } for ytd in voice_player.playing.contents]
            }
        else:
            raise CommandError(f"Non so come visualizzare il contenuto di un "
                               f"[c]{voice_player.playing.__class__.__qualname__}[/c].")



import asyncio
import logging
import sentry_sdk
import aiohttp
from typing import *
from royalnet.commands import *
from royalnet.utils import *
from royalnet.serf.telegram.escape import escape as tg_escape
from ..tables import Steam, Brawlhalla
from ..types import BrawlhallaRank, BrawlhallaMetal, BrawlhallaTier

log = logging.getLogger(__name__)


class BrawlhallaCommand(Command):
    name: str = "brawlhalla"

    aliases = ["bh", "bruhalla", "bruhlalla"]

    description: str = "Visualizza le tue statistiche di Dota!"

    syntax: str = ""

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        if self.interface.name == "telegram":
            self.loop.create_task(self._updater(900))

    async def _send(self, message):
        client = self.serf.client
        await self.serf.api_call(client.send_message,
                                 chat_id=self.config["Telegram"]["main_group_id"],
                                 text=tg_escape(message),
                                 parse_mode="HTML",
                                 disable_webpage_preview=True)

    @staticmethod
    def _display(bh: Brawlhalla) -> str:
        string = f"ℹ️ [b]{bh.name}[/b]\n\n"

        if bh.rank_1v1:
            string += f"1v1: [b]{bh.rank_1v1}[/b]\n"

        return string

    async def _notify(self,
                      obj: Brawlhalla,
                      attribute_name: str,
                      old_value: Any,
                      new_value: Any):
        if attribute_name == "rank_1v1":
            old_rank: Optional[BrawlhallaRank] = old_value
            new_rank: Optional[BrawlhallaRank] = new_value
            if new_rank > old_rank:
                message = f"📈 [b]{obj.steam.user}[/b] è salito a [b]{new_value}[/b] ({obj.rating_1v1} MMR) in 1v1 su Brawlhalla! Congratulazioni!"
            elif new_rank < old_rank:
                message = f"📉 [b]{obj.steam.user}[/b] è sceso a [b]{new_value}[/b] ({obj.rating_1v1} MMR) in 1v1 su Brawlhalla."
            else:
                return
            await self._send(message)

    @staticmethod
    async def _change(obj: Brawlhalla,
                      attribute_name: str,
                      new_value: Any,
                      callback: Callable[[Brawlhalla, str, Any, Any], Awaitable[None]]):
        old_value = obj.__getattribute__(attribute_name)
        if old_value != new_value:
            await callback(obj, attribute_name, old_value, new_value)
        obj.__setattr__(attribute_name, new_value)

    async def _update(self, steam: Steam, db_session):
        BrawlhallaT = self.alchemy.get(Brawlhalla)
        log.info(f"Updating: {steam}")
        async with aiohttp.ClientSession() as session:
            bh: Brawlhalla = steam.brawlhalla
            if bh is None:
                log.debug(f"Checking if player has an account...")
                async with session.get(f"https://api.brawlhalla.com/search?steamid={steam.steamid.as_64}&api_key={self.config['Brawlhalla']['api_key']}") as response:
                    if response.status != 200:
                        raise ExternalError(f"Brawlhalla API /search returned {response.status}!")
                    j = await response.json()
                    if j == {} or j == []:
                        log.debug("No account found.")
                        return
                    bh = BrawlhallaT(
                        steam=steam,
                        brawlhalla_id=j["brawlhalla_id"],
                        name=j["name"]
                    )
                    db_session.add(bh)
                    message = f"↔️ Account {bh} connesso a {bh.steam.user}!"
                    await self._send(message)
            async with session.get(f"https://api.brawlhalla.com/player/{bh.brawlhalla_id}/ranked?api_key={self.config['Brawlhalla']['api_key']}") as response:
                if response.status != 200:
                    raise ExternalError(f"Brawlhalla API /ranked returned {response.status}!")
                j = await response.json()
                if j == {} or j == []:
                    log.debug("No ranked info found.")
                else:
                    await self._change(bh, "rating_1v1", j["rating"], self._notify)
                    metal_name, tier_name = j["tier"].split(" ", 1)
                    metal = BrawlhallaMetal[metal_name.upper()]
                    tier = BrawlhallaTier(int(tier_name))
                    rank = BrawlhallaRank(metal=metal, tier=tier)
                    await self._change(bh, "rank_1v1", rank, self._notify)
            await asyncify(db_session.commit)

    async def _updater(self, period: int):
        log.info(f"Started updater with {period}s period")
        while True:
            log.info(f"Updating...")
            session = self.alchemy.Session()
            log.info("")
            steams = session.query(self.alchemy.get(Steam)).all()
            for steam in steams:
                try:
                    await self._update(steam, session)
                except Exception as e:
                    sentry_sdk.capture_exception(e)
                    log.error(f"Error while updating {steam.user.username}: {e}")
                await asyncio.sleep(1)
            await asyncify(session.commit)
            session.close()
            log.info(f"Sleeping for {period}s")
            await asyncio.sleep(period)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        author = await data.get_author(error_if_none=True)

        found_something = False

        message = ""
        for steam in author.steam:
            await self._update(steam, data.session)
            if steam.brawlhalla is None:
                continue
            found_something = True
            message += self._display(steam.brawlhalla)
            message += "\n"
        if not found_something:
            raise UserError("Nessun account di Brawlhalla trovato.")
        await data.reply(message)



from sqlalchemy import *
from sqlalchemy.orm import *
from sqlalchemy.ext.declarative import declared_attr
import steam
from ..types import BrawlhallaRank, BrawlhallaTier, BrawlhallaMetal


# noinspection PyAttributeOutsideInit
class Brawlhalla:
    __tablename__ = "brawlhalla"

    @declared_attr
    def brawlhalla_id(self):
        return Column(Integer, primary_key=True)

    @declared_attr
    def _steamid(self):
        return Column(BigInteger, ForeignKey("steam._steamid"), primary_key=True)

    @declared_attr
    def steam(self):
        return relationship("Steam", backref=backref("brawlhalla", uselist=False))

    @property
    def steamid(self):
        return steam.SteamID(self._steamid)

    @declared_attr
    def name(self):
        return Column(String, nullable=False)

    @declared_attr
    def rating_1v1(self):
        return Column(Integer)

    @declared_attr
    def tier_1v1(self):
        return Column(Enum(BrawlhallaTier))

    @declared_attr
    def metal_1v1(self):
        return Column(Enum(BrawlhallaMetal))

    @property
    def rank_1v1(self):
        return BrawlhallaRank(metal=self.metal_1v1, tier=self.tier_1v1)

    @rank_1v1.setter
    def rank_1v1(self, value):
        if not isinstance(value, BrawlhallaRank):
            raise TypeError("rank_1v1 can only be set to BrawlhallaRank values.")
        self.metal_1v1 = value.metal
        self.tier_1v1 = value.tier

    def __repr__(self):
        return f"<Brawlhalla account {self._steamid}>"

    def __str__(self):
        return f"[c]brawlhalla:{self.brawlhalla_id}[/c]"